package x.y.z;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.List;

import javax.jdo.annotations.Index;
import javax.jdo.annotations.Unique;
import javax.persistence.Column;
import javax.persistence.Embeddable;
import javax.persistence.Entity;
import javax.persistence.EntityManager;
import javax.persistence.EntityManagerFactory;
import javax.persistence.GeneratedValue;
import javax.persistence.GenerationType;
import javax.persistence.Id;
import javax.persistence.Lob;
import javax.persistence.Persistence;
import javax.persistence.SequenceGenerator;
import javax.persistence.TypedQuery;
import javax.persistence.criteria.CriteriaBuilder;
import javax.persistence.criteria.CriteriaQuery;
import javax.persistence.criteria.ParameterExpression;
import javax.persistence.criteria.Path;
import javax.persistence.criteria.Root;
import javax.validation.constraints.NotNull;
import javax.validation.constraints.Size;

import com.objectdb.Enhancer;


public class DBTestElementCollection {
	
	public static void main(String[] args) throws Exception{

		enhanceClasses();
		
		EntityManagerFactory emf = Persistence.createEntityManagerFactory("test3-company.odb");
		String companyId = "test"+System.currentTimeMillis();
		Company company = new Company();
		company.setCompanyId(companyId);
		String companyName = "Test Company"+companyId;
		company.setName(companyName);
		
		Branch branchOne = new Branch();
		branchOne.setBranch("BranchOne: " + Math.random());
		company.getBranches().add(branchOne);
		
		
		Branch branchTwo = new Branch();
		branchTwo.setBranch(branchOne.getBranch());
		company.getBranches().add(branchTwo);
		
		EntityManager em = emf.createEntityManager();
		em.getTransaction().begin();
		
		em.persist(company);
		em.getTransaction().commit();
		em.clear();

		// This test works, only one result
		testFindByLikeLowerWildcardWithoutBranchWithOr(emf, companyName);
		System.out.println("\n");
		
		// The next test fails - should only return one result
		testFindByLikeLowerWildcardWithBranchWithOr(emf, companyName);

	}	
	
	/**
	 * @param emf
	 * @param companyName
	 */
	private static void testFindByLikeLowerWildcardWithoutBranchWithOr(EntityManagerFactory emf,
			String companyName) {
		EntityManager em;
		// Do fetch, expect one
		em = emf.createEntityManager();
		CriteriaBuilder criteriaBuilder = em.getCriteriaBuilder();
		CriteriaQuery<Company> query = criteriaBuilder.createQuery(Company.class);
		Root<Company> companyRoot = query.from(Company.class);
		
		ParameterExpression<String> companyNamePara = criteriaBuilder.parameter(String.class);
		ParameterExpression<Boolean> conditionPara = criteriaBuilder.parameter(Boolean.class);
		Path<String> namePath = companyRoot.get("name");
		query.select(companyRoot);
		query.where(
				criteriaBuilder.like(criteriaBuilder.lower(namePath), companyNamePara),
				criteriaBuilder.and(
						criteriaBuilder.or(
						criteriaBuilder.equal(companyRoot.get("conditionOne"), conditionPara)
				)
				)
		);
		
		System.out.println("testFindByLikeLowerWildcardWithoutBranchWithOr: " + query);
		TypedQuery<Company> typedQuery = em.createQuery(query);
		typedQuery.setParameter(companyNamePara, "%"+new String(companyName).toLowerCase()+"%");
		typedQuery.setParameter(conditionPara, false);
		
		
		List<Company> resultList = typedQuery.getResultList();
		if(resultList.size() > 1){
			System.err.println("Only expected one result, but got: " + resultList.size());
		}else{
			System.out.println("Got: " + resultList.size() +" results - correct");
		}
		
		if(resultList.size() == 0){
			
		}
		em.close();
	}	
	
	/**
	 * @param emf
	 * @param companyName
	 */
	private static void testFindByLikeLowerWildcardWithBranchWithOr(EntityManagerFactory emf,
			String companyName) {
		EntityManager em;
		// Do fetch, expect one
		em = emf.createEntityManager();
		CriteriaBuilder criteriaBuilder = em.getCriteriaBuilder();
		CriteriaQuery<Company> query = criteriaBuilder.createQuery(Company.class);
		Root<Company> companyRoot = query.from(Company.class);
		
		ParameterExpression<String> companyNamePara = criteriaBuilder.parameter(String.class);
		ParameterExpression<String> branchNamePara = criteriaBuilder.parameter(String.class);
		ParameterExpression<Boolean> conditionPara = criteriaBuilder.parameter(Boolean.class);
		Path<String> namePath = companyRoot.get("name");
		query.select(companyRoot);
		query.where(
				criteriaBuilder.like(criteriaBuilder.lower(namePath), companyNamePara),
				criteriaBuilder.and(
						criteriaBuilder.or(
						criteriaBuilder.equal(companyRoot.get("branches").get("branch"), branchNamePara),
						criteriaBuilder.equal(companyRoot.get("conditionOne"), conditionPara)
				)
				)
		);
		
		System.out.println("testFindByLikeLowerWildcardWithBranchWithOr: " + query);
		TypedQuery<Company> typedQuery = em.createQuery(query);
		typedQuery.setParameter(companyNamePara, "%"+new String(companyName).toLowerCase()+"%");
		// This name doesn't exist, but it doesn't matter - we are doing an 'or' - which means that 
		// even if no match is found for branch, the company will still be returned for other conditions that match.
		// Adding this returns the company twice.
		typedQuery.setParameter(branchNamePara, "any old name, still matches");
		typedQuery.setParameter(conditionPara, false);
		
		
		List<Company> resultList = typedQuery.getResultList();
		if(resultList.size() > 1){
			System.err.println("Only expected one result, but got: " + resultList.size());
		}else{
			System.out.println("Got: " + resultList.size() +" results - correct");
		}
		
		if(resultList.size() == 0){
			
		}
		em.close();
	}	
		
	
	/**
	 * 
	 */
	private static void enhanceClasses() {
		Enhancer.enhance(Company.class.getName()+"," +Branch.class.getName());
	}
	
	@Entity
	@SequenceGenerator(name="company_seq", initialValue=1,allocationSize=5)
	public static class Company {
		
	    @Id 
	    @GeneratedValue(strategy=GenerationType.IDENTITY, generator="company_seq")
	    private Long id;		

	    /**
		 * 
		 */
		private static final long serialVersionUID = -299617781352746768L;

		/**
	     * Name
	     */
	    @Column(nullable = false, length = 128)
	    @Unique(name="uq_company_name")
	    private String name;
	    
	    private boolean conditionOne = false;
	    
	    /**
	     * Short name to identify company
	     */
	    @Column(name="company_id")
	    @NotNull
	    @Size(min=3)
	    @Index(name="idx_company_id")
	    @Unique(name="uq_company_id")
	    private String companyId;

	    @Column(name = "detailed_desc")
	    @Lob
	    private String detailedDescription;
	    
	    private List<Branch> branches = new ArrayList<DBTestElementCollection.Branch>();

	    public String getName() {
	        return name;
	    }

	    public void setName(String name) {
	        this.name = name;
	    }

	    public String getDetailedDescription() {
	        return detailedDescription;
	    }

	    public void setDetailedDescription(String detailedDescription) {
	        this.detailedDescription = detailedDescription;
	    }

		/**
		 * @return the id
		 */
		public Long getId() {
			return id;
		}

		/**
		 * @return the companyId
		 */
		public String getCompanyId() {
			return companyId;
		}

		/**
		 * @param companyId the companyId to set
		 */
		public void setCompanyId(String companyId) {
			this.companyId = companyId;
		}

		/**
		 * @return the branches
		 */
		public List<Branch> getBranches() {
			return branches;
		}

		/**
		 * @param branches the branches to set
		 */
		public void setBranches(List<Branch> branches) {
			this.branches = branches;
		}

		/**
		 * @return the conditionOne
		 */
		public boolean isConditionOne() {
			return conditionOne;
		}

		/**
		 * @param conditionOne the conditionOne to set
		 */
		public void setConditionOne(boolean conditionOne) {
			this.conditionOne = conditionOne;
		}
	}
	
	@Embeddable
	public static class Branch implements Serializable{

		/**
		 * 
		 */
		private static final long serialVersionUID = 4038843390715206871L;
		
		private String branch;

		/**
		 * @return the name
		 */
		public String getBranch() {
			return branch;
		}

		/**
		 * @param name the name to set
		 */
		public void setBranch(String name) {
			this.branch = name;
		}

		/* (non-Javadoc)
		 * @see java.lang.Object#hashCode()
		 */
		@Override
		public int hashCode() {
			final int prime = 31;
			int result = 1;
			result = prime * result + ((branch == null) ? 0 : branch.hashCode());
			return result;
		}

		/* (non-Javadoc)
		 * @see java.lang.Object#equals(java.lang.Object)
		 */
		@Override
		public boolean equals(Object obj) {
			if (this == obj)
				return true;
			if (obj == null)
				return false;
			if (getClass() != obj.getClass())
				return false;
			Branch other = (Branch) obj;
			if (branch == null) {
				if (other.branch != null)
					return false;
			} else if (!branch.equals(other.branch))
				return false;
			return true;
		}
	}
}